/*
     []---------------------------------------------------------------[]
     |                           GRAF3D.C                              |
     |                                                                 |
     |            Construcao do grafico de duas superficies            |
     |                                                                 |
     |            por Lenimar N. Andrade, maio/95--maio/98             |
     []---------------------------------------------------------------[]
*/

#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <graphics.h>
#include <dir.h>

typedef struct { double xmin, xmax, ymin, ymax; } retangulo;
#define Dominio retangulo

#define PI		3.14159265358979323846
#define PI_SOBRE_2	1.57079632679489661923
#define PI_SOBRE_4      0.78539816339744830961
#define DOIS_PI         6.28318530717958647692

#define SQR(x)  ((x)*(x))
#define DISTANCIA(a, b, c, x, y, z) sqrt(SQR(a-x) + SQR(b-y) + SQR(c-z))
#define PROD_VET(V1, V2, V3, a1, a2, a3, b1, b2, b3) V1 = a2*b3-a3*b2; \
				   V2 = a3*b1-a1*b3; V3 = a1*b2-a2*b1
#define NORMA(V1, V2, V3)  (sqrt((V1)*(V1) + (V2)*(V2) + (V3)*(V3)))
double FI = -0.3, TETA = -0.7;

#define GRAF

double dist_obs  = 10000,   /* Distancia do observador `a origem           */
      X_CENTRO = 320, Y_CENTRO = 240, /* Coordenadas do centro da tela     */
      TAM = 50,                   /* Coeficiente de ampliacao              */
      TAM_EIXOS = 3,              /* Tamanho dos eixos coordenados         */
      n1, n2, n3,                 /* Vetor normal                          */
      x0, y0, z0,                 /* Posicao do observador                 */
      nuF = 30, nvF = 30,         /* Subdivisoes do dominio                */
      nuG = 30, nvG = 30,         /* Subdivisoes do dominio                */
      inc = 0.02,                 /* Incremento do parametro               */
      o1 = 1, o2 = 1, o3 = 1;     /* Direcao do observador                 */
  /*  o1 = cos(TETA)*cos(FI), o2 = -sin(TETA)*cos(FI), o3 = -sin(FI);      */

extern unsigned _stklen = 20000;  /* Ajusta tamanho da pilha */

/* ------------------------------------------------------------------------- */

void ProjetaPonto(double x, double y, double z,
					 double *proj_X, double *proj_Y) {

/* Projeta o ponto (x, y, z) em coordenadas (proj_X, proj_Y) da tela */

  *proj_X = y*cos(TETA) + x*sin(TETA);
  *proj_Y = z*cos(FI) + x*cos(TETA)*sin(FI) - y*sin(FI)*sin(TETA);
}

/* ------------------------------------------------------------------------- */

int IniciaModoGrafico(void) {

/* Inicializa o modo grafico. Retorna 1 no caso de sucesso ou 0 em caso
   contrario. */

  int gm, gd = DETECT;

  registerbgidriver(EGAVGA_driver);
  initgraph(&gd, &gm, "");
  return (graphresult() == grOk);
}

/* ------------------------------------------------------------------------ */

int VetorNormal(double *N0, double *N1, double *N2, double (*f1)(double, double),
		double (*f2)(double, double), double (*f3)(double, double),
		double u0, double v0) {

/* Calcula (*N0, *N1, *N2) o vetor normal no  ponto  (u0, v0)  `a
   superficie f(u, v) = (f1(u, v), f2(u, v), f3(u, v)). Retorna 0
   se o vetor normal encontrado for nulo ou 1 em  caso contrario. */

  const double h = 1E-6;
  double a1, a2, a3, b1, b2, b3, norma_N;
  const double epsilon = 1e-6;

  /* Calculo de 2 vetores tangentes `a superficie f */

  a1 = (f1(u0 + h, v0) - f1(u0 - h, v0))/(2.0*h);
  a2 = (f2(u0 + h, v0) - f2(u0 - h, v0))/(2.0*h);
  a3 = (f3(u0 + h, v0) - f3(u0 - h, v0))/(2.0*h);

  b1 = (f1(u0, v0 + h) - f1(u0, v0 - h))/(2.0*h);
  b2 = (f2(u0, v0 + h) - f2(u0, v0 - h))/(2.0*h);
  b3 = (f3(u0, v0 + h) - f3(u0, v0 - h))/(2.0*h);

  /* Vetor normal a f */
  PROD_VET(*N0, *N1, *N2, a1, a2, a3, b1, b2, b3);

  norma_N = NORMA(*N0, *N1, *N2);
  if (norma_N < epsilon) return 0;
  return 1;
}

/* ------------------------------------------------------------------------- */

void DesenhaEixos(int tam, int cor) {

/* Desenha os eixos x, y, z usando tamanho e cor especificados */

  double X0, Y0, X, Y;
  int cor_aux, x_x, x_y, y_x, y_y, z_x, z_y;

  cor_aux = getcolor();
  setcolor(cor);
  ProjetaPonto(0, 0, 0, &X0, &Y0);
  ProjetaPonto(tam, 0, 0, &X, &Y);
  line((int) (X_CENTRO + X0*TAM), (int) (Y_CENTRO - Y0*TAM),
       (int) (X_CENTRO + X*TAM), (int) (Y_CENTRO - Y*TAM));
  x_x = (int) (X_CENTRO + X*TAM); x_y = (int) (Y_CENTRO - Y*TAM);
  ProjetaPonto(0, tam, 0, &X, &Y);
  line((int) (X_CENTRO + X0*TAM), (int) (Y_CENTRO - Y0*TAM),
       (int) (X_CENTRO + X*TAM), (int) (Y_CENTRO - Y*TAM));
  y_x = (int) (X_CENTRO + X*TAM); y_y = (int) (Y_CENTRO - Y*TAM);
  ProjetaPonto(0, 0, tam, &X, &Y);
  line((int) (X_CENTRO + X0*TAM), (int) (Y_CENTRO - Y0*TAM),
       (int) (X_CENTRO + X*TAM), (int) (Y_CENTRO - Y*TAM));
  z_x = (int) (X_CENTRO + X*TAM); z_y = (int) (Y_CENTRO - Y*TAM);

  setcolor(cor_aux);
  outtextxy(x_x, x_y, "x");
  outtextxy(y_x, y_y, "y");
  outtextxy(z_x, z_y, "z");
}

/* ------------------------------------------------------------------------- */

void DesenhaPonto(double x, double y, double z) {

/* Desenha o ponto (x, y, z) usando a cor atual. */

  double X, Y;

  X = y*cos(TETA) + x*sin(TETA);
  Y = z*cos(FI) + x*cos(TETA)*sin(FI) - y*sin(FI)*sin(TETA);
  putpixel((int) (X_CENTRO + X*TAM), (int) (Y_CENTRO - Y*TAM), getcolor());
}

/* ------------------------------------------------------------------------- */

void DesenhaPontoCor(double x, double y, double z, int cor) {

/* Desenha o ponto (x, y, z) usando a cor especificada. */

  double X, Y;

  X = y*cos(TETA) + x*sin(TETA);
  Y = z*cos(FI) + x*cos(TETA)*sin(FI) - y*sin(FI)*sin(TETA);
  putpixel((int) (X_CENTRO + X*TAM), (int) (Y_CENTRO - Y*TAM), cor);
}

/* ------------------------------------------------------------------------- */

int InterceptaReta( long a1, long b1, long a2, long b2,
		    long a3, long b3, long a4, long b4) {

/* Verifica se o segmento de reta (a1, b1) - (a2, b2) intercepta o segmento
   (a3, b3) - (a4, b4)  em  um  unico  ponto.   Nao   sao  consideradas  as
   extremidades dos segmentos. */

  double det, t, s;

  det = (a2 - a1)*(b3 - b4) - (a3 - a4)*(b2 - b1);
  if (det == 0) return 0;

  t = ((a3 - a1)*(b3 - b4) - (a3 - a4)*(b3 - b1))/det;
  s = ((a2 - a1)*(b3 - b1) - (a3 - a1)*(b2 - b1))/det;

  return (0 < t && t < 1 && 0 < s && s < 1);

}

/* ------------------------------------------------------------------------- */

int FechoConvexo(int *a1, int *b1, int *a2, int *b2,
		  int *a3, int *b3, int *a4, int *b4) {

/* Determina o fecho convexo do poligono
		(a1, b1) - (a2, b2) - (a3, b3) - (a4, b4)          */

  int a_aux, b_aux;

  if (InterceptaReta(*a1, *b1, *a2, *b2, *a3, *b3, *a4, *b4)) {
    a_aux = *a1; b_aux = *b1;
    *a1 = *a4;   *b1 = *b4;
    *a4 = a_aux; *b4 = b_aux;
    return 14;
  }

  if (InterceptaReta(*a1, *b1, *a4, *b4, *a3, *b3, *a2, *b2)) {
    a_aux = *a1; b_aux = *b1;
    *a1 = *a2;   *b1 = *b2;
    *a2 = a_aux; *b2 = b_aux;
    return 12;
  }

  return 0;
}

/* ------------------------------------------------------------------------- */

void DesenhaReta(double a1, double a2, double a3,
		 double b1, double b2, double b3, int cor) {

/* Desenha a reta ligando os pontos (a1, a2, a3) - (b1, b2, b3) */

  double x1, x2, y1, y2;

  setcolor(cor);
  ProjetaPonto(a1, a2, a3, &x1, &y1);
  ProjetaPonto(b1, b2, b3, &x2, &y2);
  line((int) (X_CENTRO + TAM*x1), (int) (Y_CENTRO - TAM*y1),
       (int) (X_CENTRO + TAM*x2), (int) (Y_CENTRO - TAM*y2));
}

/* ------------------------------------------------------------------------- */

typedef struct {int x, y; } t_ponto;

#define MAX_PONTOS 3000
t_ponto p[MAX_PONTOS];
int quant_pontos = 0;

void SequenciaDePontosDaReta(int x0, int y0, int x1, int y1) {

/* Determina os pontos da reta (x0, y0) - (x1, y1) usando o algoritmo de
   Bresenham. Os pontos sao guardados na matriz de pontos global p[]. */

  int aux, primeiro_octante, dx, dy, incrE, incrNE,
      X0, X1, Y0, Y1, d, x, y, inclinacao_negativa;

  if (x0 > x1) {
    aux = x0; x0 = x1; x1 = aux;
    aux = y0; y0 = y1; y1 = aux;
  }

  X0 = x0; X1= x1; Y0 = y0; Y1 = y1;

  inclinacao_negativa = (y0 > y1);
  if (inclinacao_negativa) {
    aux = y0; y0 = y1; y1 = aux;
  }

  primeiro_octante = (y1 - y0 <= x1 - x0);
  if (!primeiro_octante) {
    aux = x0; x0 = y0; y0 = aux;
    aux = x1; x1 = y1; y1 = aux;
  }

  dx = x1 - x0;
  dy = y1 - y0;
  d = 2*dy - dx;
  incrE = 2*dy;
  incrNE = 2*(dy - dx);
  x = x0;
  y = y0;

  if (primeiro_octante)
    if (inclinacao_negativa)
      { p[quant_pontos].x = x; p[quant_pontos].y = Y0 + Y1 - y; }
    else
      { p[quant_pontos].x = x; p[quant_pontos].y = y; }
  else
    if (inclinacao_negativa)
      { p[quant_pontos].x = X0 + X1 - y; p[quant_pontos].y = x; }
    else
      { p[quant_pontos].x = y; p[quant_pontos].y = x; }
  quant_pontos++;

  while (x < x1) {

    if (d < 0) {
      d = d + incrE;
      x = x + 1;
    } else {
      d = d + incrNE;
      x = x + 1;
      y = y + 1;
    }

    if (primeiro_octante)
      if (inclinacao_negativa)
	{ p[quant_pontos].x = x; p[quant_pontos].y = Y0 + Y1 - y; }
      else
	{ p[quant_pontos].x = x; p[quant_pontos].y = y; }
    else
      if (inclinacao_negativa)
	{ p[quant_pontos].x = X0 + X1 - y; p[quant_pontos].y = x; }
      else
	{ p[quant_pontos].x = y; p[quant_pontos].y = x; }
    quant_pontos++;
  }
}

/* ----------------------------------------------------------------------- */

void Preenche(int cor, int cor_secundaria, int padrao) {

/* Preenche a regiao delitada pela matriz de pontos p[]. Sao usadas duas
   cores de preenchimento misturadas segundo o padrao de bits indicado. */

  int teste, cont = 0, i, a0, b0, a1, b1, m, M, linha, col;

  a0 = 10000; a1 = -10000;
  b0 = 10000; b1 = -10000;
  for (i = 0; i < quant_pontos; i++) {
   if (a0 > p[i].x) a0 = p[i].x;
   if (b0 > p[i].y) b0 = p[i].y;
   if (a1 < p[i].x) a1 = p[i].x;
   if (b1 < p[i].y) b1 = p[i].y;
  }

  for (linha = b0; linha <= b1; linha++) {
    m = 10000; M = -10000;
    for (i = 0; i < quant_pontos; i++)
      if (linha == p[i].y) {
	if (m > p[i].x) m = p[i].x;
	if (M < p[i].x) M = p[i].x;
      }
    for (col = m; col <= M; col++) {
      cont++;
      cont %= 8;
      teste = padrao >> cont;
      if (teste & 1)
	putpixel(col, linha, cor);
      else
	putpixel(col, linha, cor_secundaria);
    }
  }
}

/* ----------------------------------------------------------------------- */

void DeterminaPontos(int a, int b, int c, int d, int e, int f) {

  quant_pontos = 0;
  SequenciaDePontosDaReta(a, b, c, d);
  SequenciaDePontosDaReta(c, d, e, f);
  SequenciaDePontosDaReta(a, b, e, f);
  if (quant_pontos > MAX_PONTOS) return;
}

/* ----------------------------------------------------------------------- */

void Pintar(int a, int b, int c, int d, int e, int f, int g, int h,
	    int cor_front, int cor_int1, int cor_int2, int padrao) {

/* Pinta a regiao (a, b) - (c, d) - (e, f) - (g, h). A fronteira e' pintada
   com  a cor cor_front. O interior  e' pintado  com  as cores  cor_int1  e
   cor_int2 misturadas conforme os bits do parametro padrao. */

  DeterminaPontos(a, b, c, d, e, f);
  Preenche(cor_int1, cor_int2, padrao);
  DeterminaPontos(a, b, e, f, g, h);
  Preenche(cor_int1, cor_int2, padrao);
  setcolor(cor_front);
  line(a, b, c, d);
  line(c, d, e, f);
  line(e, f, g, h);
  line(g, h, a, b);
}

/* ----------------------------------------------------------------------- */

#define D_OBS(X, Y, Z)  sqrt(SQR(X - o1*dist_obs) + SQR(Y - o2*dist_obs) \
			   + SQR(Z - o3*dist_obs))

double DistanciaObs(double xp1, double yp1, double zp1,
		   double xp2, double yp2, double zp2,
		   double xp3, double yp3, double zp3,
		   double xp4, double yp4, double zp4) {

/* Calcula a distancia media do observador `a regiao (xp1, yp1, zp1) -
   (xp2, yp2, zp2) - (xp3, yp3, zp3) - (xp4, yp4, zp4) */
  double d1, d2, d3, d4, D1, D2, D3;

  D1 = o1*dist_obs; D2 = o2*dist_obs; D3 = o3*dist_obs;
  d1 =  sqrt(SQR(xp1 - D1) + SQR(yp1 - D2) + SQR(zp1 - D3));
  d2 =  sqrt(SQR(xp2 - D1) + SQR(yp2 - D2) + SQR(zp2 - D3));
  d3 =  sqrt(SQR(xp3 - D1) + SQR(yp3 - D2) + SQR(zp3 - D3));
  d4 =  sqrt(SQR(xp4 - D1) + SQR(yp4 - D2) + SQR(zp4 - D3));

  return (d1 + d2 + d3 + d4)/4.0;
}

/* ------------------------------------------------------------------------- */

typedef struct {
  double d;                  /* Distancia do observador `a media dos vertices */
  int xq1, yq1, xq2, yq2, xq3, yq3, xq4, yq4;  /* Vertices                   */
  double cosN; /* Cosseno do angulo entre os vetores normal e o do observador */
  int n;                    /* Numero do grafico                             */
} t_regiao;

/* ------------------------------------------------------------------------- */

void LeRegiao(FILE *arq, long n, t_regiao *Regiao) {

/* Le regiao no arquivo arq. A primeira regiao e' a de ordem 0. */

  fseek(arq, sizeof(t_regiao)*n, SEEK_SET);
  fread(Regiao, sizeof(t_regiao), 1, arq);
};

/* ------------------------------------------------------------------------- */

long QuantasRegioes(char *nome_arq) {

/* Determina quantas regioes foram gravadas no arquivo com nome nome_arq */

  struct ffblk x;

  findfirst(nome_arq, &x, 0);
  return x.ff_fsize/sizeof(t_regiao);
}

/* ------------------------------------------------------------------------- */

double LeDistanciaEmArquivo(FILE *arq, long rec) {

  t_regiao Regiao;

  fseek(arq, rec*sizeof(t_regiao), SEEK_SET);
  if (fread(&Regiao, sizeof(t_regiao), 1, arq))
    return Regiao.d;
  else
    return NULL;
}

/* ------------------------------------------------------------------------- */

void TrocaItensEmArquivo(FILE *arq, long i, long j) {

  /* Troca os itens de ordens i e j no arquivo arq */

  char a[sizeof(t_regiao)], b[sizeof(t_regiao)];

  fseek(arq, sizeof(t_regiao)*i, SEEK_SET);
  fread(a, sizeof(t_regiao), 1, arq);
  fseek(arq, sizeof(t_regiao)*j, SEEK_SET);
  fread(b, sizeof(t_regiao), 1, arq);

  fseek(arq, sizeof(t_regiao)*j, SEEK_SET);
  fwrite(a, sizeof(t_regiao), 1, arq);
  fseek(arq, sizeof(t_regiao)*i, SEEK_SET);
  fwrite(b, sizeof(t_regiao), 1, arq);
}

/* ------------------------------------------------------------------------- */

void qs_disco(FILE *arq, long esquerda, long direita) {

  /* Classifica o arquivo arq, colocando em ordem crescente os campos
     distancia */

  long i, j;
  double x;

  i = esquerda; j = direita;
  x = LeDistanciaEmArquivo(arq, (i + j)/2);

  do {
    while (LeDistanciaEmArquivo(arq, i) < x && i < direita) i++;
    while (LeDistanciaEmArquivo(arq, j) > x && j > esquerda) j--;
    if (i <= j) {
      TrocaItensEmArquivo(arq, i, j);
      i++; j--;
    }
  } while (i <= j);
  if (esquerda < j) qs_disco(arq, esquerda, j);
  if (i < direita) qs_disco(arq, i, direita);
}

/* ------------------------------------------------------------------------- */

void ClassificaRegioes(FILE *arq, char *nome_arq) {

  long quant;

  quant = QuantasRegioes(nome_arq);
  qs_disco(arq, 0, quant - 1);
}

/* ------------------------------------------------------------------------- */

void ArestaDoCubo(double a, double b, double c,
		  double d, double e, double f, int cor) {

  int X1, Y1, X2, Y2;
  double p1, p2;

  setcolor(cor);
  ProjetaPonto(a, b, c, &p1, &p2);
  X1 = (int) (X_CENTRO + p1*TAM);
  Y1 = (int) (Y_CENTRO - p2*TAM);
  ProjetaPonto(d, e, f, &p1, &p2);
  X2 = (int) (X_CENTRO + p1*TAM);
  Y2 = (int) (Y_CENTRO - p2*TAM);
  line(X1, Y1, X2, Y2);
}

/* ------------------------------------------------------------------------- */

int Desenha2Superficies(double (*f1) (double,double), double (*f2)(double, double),
  double (*f3)(double, double), double umin1, double umax1, double vmin1, double vmax1,
  int cor_f1_1, int cor_f2_1, int cor_int1_1, int cor_int2_1, int cor_int3_1,
  int cor_int4_1, int padrao1_1, int padrao2_1,
  double (*g1) (double,double), double (*g2)(double, double),
  double (*g3)(double, double), double umin2, double umax2, double vmin2, double vmax2,
  int cor_f1_2, int cor_f2_2, int cor_int1_2, int cor_int2_2, int cor_int3_2,
  int cor_int4_2, int padrao1_2, int padrao2_2) {

/* Desenha as superficies (f1, f2, f3) e (g1, g2, g3), apagando as curvas
   invisiveis.

   A parte superior da superficie e' pintada com  a  cor de  fronteira cor_f1 e
   cores cor_int1 e cor_int2 misturadas conforme  o  padrao  de  bits  padrao1.

   A parte inferior da superficie e' pintada com a cor de  fronteira  cor_f2  e
   cores cor_int3 e cor_int4  misturadas conforme  o  padrao  de  bits padrao2.
*/

  double u, v, xp1, yp1, zp1, xp2, yp2, zp2, xp3, yp3, zp3, xp4, yp4, zp4,
	X, Y, distancia, n1, n2, n3, xmin, xmax, ymin, ymax, zmax, zmin,
	smax, s1, s2, s3, s4, s5, s6, s7, s8;
  int  xq1, xq2, xq3, xq4, yq1, yq2, yq3, yq4;
  t_regiao re;
  long i, quant;
  FILE *arq;
  char arq_dados[20] = "PONTOS.DAT";

  int opcao = -1;    /* Se  opcao  for  maior  ou igual  a  0,  e'
			desenhado  um  cubo envolvendo a superficie
			desenhada usando a cor = opcao. */

  o1 = cos(TETA)*cos(FI), o2 = -sin(TETA)*cos(FI), o3 = -sin(FI);

  if ((arq = fopen(arq_dados, "wb")) == NULL) return -1;

  xmax = xmin = f1(umin1, vmin1);
  ymax = ymin = f2(umin1, vmin1);
  zmax = zmin = f3(umin1, vmin1);

  for (u = umin1 + (umax1 - umin1)/nuF; u <= umax1 + (umax1 - umin1)/(2*nuF);
					       u += (umax1 - umin1)/nuF)
    for (v = vmin1 + (vmax1 - vmin1)/nvF; v <= vmax1 + (vmax1 - vmin1)/(2*nvF);
					       v += (vmax1 - vmin1)/nvF) {
      xp1 = f1(u, v);
      yp1 = f2(u, v);
      zp1 = f3(u, v);
      if (xmax < xp1) xmax = xp1;
      if (xmin > xp1) xmin = xp1;
      if (ymax < yp1) ymax = yp1;
      if (ymin > yp1) ymin = yp1;
      if (zmax < zp1) zmax = zp1;
      if (zmin > zp1) zmin = zp1;

      xp2 = f1(u - (umax1 - umin1)/nuF, v);
      yp2 = f2(u - (umax1 - umin1)/nuF, v);
      zp2 = f3(u - (umax1 - umin1)/nuF, v);
      if (xmax < xp2) xmax = xp2;
      if (xmin > xp2) xmin = xp2;
      if (ymax < yp2) ymax = yp2;
      if (ymin > yp2) ymin = yp2;
      if (zmax < zp2) zmax = zp2;
      if (zmin > zp2) zmin = zp2;

      xp3 = f1(u, v - (vmax1 - vmin1)/nvF);
      yp3 = f2(u, v - (vmax1 - vmin1)/nvF);
      zp3 = f3(u, v - (vmax1 - vmin1)/nvF);
      if (xmax < xp3) xmax = xp3;
      if (xmin > xp3) xmin = xp3;
      if (ymax < yp3) ymax = yp3;
      if (ymin > yp3) ymin = yp3;
      if (zmax < zp3) zmax = zp3;
      if (zmin > zp3) zmin = zp3;

      xp4 = f1(u - (umax1 - umin1)/nuF, v - (vmax1 - vmin1)/nvF);
      yp4 = f2(u - (umax1 - umin1)/nuF, v - (vmax1 - vmin1)/nvF);
      zp4 = f3(u - (umax1 - umin1)/nuF, v - (vmax1 - vmin1)/nvF);
      if (xmax < xp4) xmax = xp4;
      if (xmin > xp4) xmin = xp4;
      if (ymax < yp4) ymax = yp4;
      if (ymin > yp4) ymin = yp4;
      if (zmax < zp4) zmax = zp4;
      if (zmin > zp4) zmin = zp4;

      ProjetaPonto(xp1, yp1, zp1, &X, &Y);
      xq1 = (int) (X_CENTRO + X*TAM); yq1 = (int) (Y_CENTRO - Y*TAM);
      ProjetaPonto(xp2, yp2, zp2, &X, &Y);
      xq2 = (int) (X_CENTRO + X*TAM); yq2 = (int) (Y_CENTRO - Y*TAM);
      ProjetaPonto(xp3, yp3, zp3, &X, &Y);
      xq3 = (int) (X_CENTRO + X*TAM); yq3 = (int) (Y_CENTRO - Y*TAM);
      ProjetaPonto(xp4, yp4, zp4, &X, &Y);
      xq4 = (int) (X_CENTRO + X*TAM); yq4 = (int) (Y_CENTRO - Y*TAM);

      FechoConvexo(&xq1, &yq1, &xq2, &yq2, &xq3, &yq3, &xq4, &yq4);
      distancia = DistanciaObs(xp1, yp1, zp1, xp2, yp2, zp2,
			       xp3, yp3, zp3, xp4, yp4, zp4);
      VetorNormal(&n1, &n2, &n3, f1, f2, f3, u - (umax1 - umin1)/(2*nuF),
					     v - (vmax1 - vmin1)/(2*nvF));
      re.cosN = (n1*o1 + n2*o2 + n3*o3)/NORMA(n1, n2, n3);

      re.d = distancia;
      re.xq1 = xq1;
      re.yq1 = yq1;
      re.xq2 = xq2;
      re.yq2 = yq2;
      re.xq3 = xq3;
      re.yq3 = yq3;
      re.xq4 = xq4;
      re.yq4 = yq4;
      re.n = 1;
      fwrite(&re, sizeof(t_regiao), 1, arq);
    }

  for (u = umin2 + (umax2 - umin2)/nuG; u <= umax2 + (umax2 - umin2)/(2*nuG);
					       u += (umax2 - umin2)/nuG)
    for (v = vmin2 + (vmax2 - vmin2)/nvG; v <= vmax2 + (vmax2 - vmin2)/(2*nvG);
					       v += (vmax2 - vmin2)/nvG) {
      xp1 = g1(u, v);
      yp1 = g2(u, v);
      zp1 = g3(u, v);
      if (xmax < xp1) xmax = xp1;
      if (xmin > xp1) xmin = xp1;
      if (ymax < yp1) ymax = yp1;
      if (ymin > yp1) ymin = yp1;
      if (zmax < zp1) zmax = zp1;
      if (zmin > zp1) zmin = zp1;

      xp2 = g1(u - (umax2 - umin2)/nuG, v);
      yp2 = g2(u - (umax2 - umin2)/nuG, v);
      zp2 = g3(u - (umax2 - umin2)/nuG, v);
      if (xmax < xp2) xmax = xp2;
      if (xmin > xp2) xmin = xp2;
      if (ymax < yp2) ymax = yp2;
      if (ymin > yp2) ymin = yp2;
      if (zmax < zp2) zmax = zp2;
      if (zmin > zp2) zmin = zp2;

      xp3 = g1(u, v - (vmax2 - vmin2)/nvG);
      yp3 = g2(u, v - (vmax2 - vmin2)/nvG);
      zp3 = g3(u, v - (vmax2 - vmin2)/nvG);
      if (xmax < xp3) xmax = xp3;
      if (xmin > xp3) xmin = xp3;
      if (ymax < yp3) ymax = yp3;
      if (ymin > yp3) ymin = yp3;
      if (zmax < zp3) zmax = zp3;
      if (zmin > zp3) zmin = zp3;

      xp4 = g1(u - (umax2 - umin2)/nuG, v - (vmax2 - vmin2)/nvG);
      yp4 = g2(u - (umax2 - umin2)/nuG, v - (vmax2 - vmin2)/nvG);
      zp4 = g3(u - (umax2 - umin2)/nuG, v - (vmax2 - vmin2)/nvG);
      if (xmax < xp4) xmax = xp4;
      if (xmin > xp4) xmin = xp4;
      if (ymax < yp4) ymax = yp4;
      if (ymin > yp4) ymin = yp4;
      if (zmax < zp4) zmax = zp4;
      if (zmin > zp4) zmin = zp4;

      ProjetaPonto(xp1, yp1, zp1, &X, &Y);
      xq1 = (int) (X_CENTRO + X*TAM); yq1 = (int) (Y_CENTRO - Y*TAM);
      ProjetaPonto(xp2, yp2, zp2, &X, &Y);
      xq2 = (int) (X_CENTRO + X*TAM); yq2 = (int) (Y_CENTRO - Y*TAM);
      ProjetaPonto(xp3, yp3, zp3, &X, &Y);
      xq3 = (int) (X_CENTRO + X*TAM); yq3 = (int) (Y_CENTRO - Y*TAM);
      ProjetaPonto(xp4, yp4, zp4, &X, &Y);
      xq4 = (int) (X_CENTRO + X*TAM); yq4 = (int) (Y_CENTRO - Y*TAM);

      FechoConvexo(&xq1, &yq1, &xq2, &yq2, &xq3, &yq3, &xq4, &yq4);
      distancia = DistanciaObs(xp1, yp1, zp1, xp2, yp2, zp2,
			       xp3, yp3, zp3, xp4, yp4, zp4);
      VetorNormal(&n1, &n2, &n3, g1, g2, g3, u - (umax2 - umin2)/(2*nuG),
					     v - (vmax2 - vmin2)/(2*nvG));
      re.cosN = (n1*o1 + n2*o2 + n3*o3)/NORMA(n1, n2, n3);

      re.d = distancia;
      re.xq1 = xq1;
      re.yq1 = yq1;
      re.xq2 = xq2;
      re.yq2 = yq2;
      re.xq3 = xq3;
      re.yq3 = yq3;
      re.xq4 = xq4;
      re.yq4 = yq4;
      re.n = 2;
      fwrite(&re, sizeof(t_regiao), 1, arq);
    }

  fclose(arq);

  if ((arq = fopen(arq_dados, "r+b")) == NULL) return -1;
  ClassificaRegioes(arq, arq_dados);

  smax = s1 = D_OBS(xmin, ymin, zmin);
  s2 = D_OBS(xmin, ymin, zmax);
  if (smax < s2) smax = s2;
  s3 = D_OBS(xmin, ymax, zmin);
  if (smax < s3) smax = s3;
  s4 = D_OBS(xmin, ymax, zmax);
  if (smax < s4) smax = s4;
  s5 = D_OBS(xmax, ymin, zmin);
  if (smax < s5) smax = s5;
  s6 = D_OBS(xmax, ymin, zmax);
  if (smax < s6) smax = s6;
  s7 = D_OBS(xmax, ymax, zmin);
  if (smax < s7) smax = s7;
  s8 = D_OBS(xmax, ymax, zmax);
  if (smax < s8) smax = s8;

  if (opcao >= 0) {
    ArestaDoCubo(xmin, ymin, zmin, xmax, ymin, zmin, opcao);
    ArestaDoCubo(xmin, ymin, zmin, xmin, ymax, zmin, opcao);
    ArestaDoCubo(xmax, ymax, zmin, xmax, ymin, zmin, opcao);
    ArestaDoCubo(xmax, ymax, zmin, xmin, ymax, zmin, opcao);

    ArestaDoCubo(xmin, ymin, zmax, xmax, ymin, zmax, opcao);
    ArestaDoCubo(xmin, ymin, zmax, xmin, ymax, zmax, opcao);
    ArestaDoCubo(xmax, ymax, zmax, xmax, ymin, zmax, opcao);
    ArestaDoCubo(xmax, ymax, zmax, xmin, ymax, zmax, opcao);

    ArestaDoCubo(xmin, ymin, zmin, xmin, ymin, zmax, opcao);
    ArestaDoCubo(xmax, ymin, zmin, xmax, ymin, zmax, opcao);
    ArestaDoCubo(xmin, ymax, zmin, xmin, ymax, zmax, opcao);
    ArestaDoCubo(xmax, ymax, zmin, xmax, ymax, zmax, opcao);
  }

  quant = QuantasRegioes(arq_dados);
  for (i = quant - 1; i >= 0; i--) {
    LeRegiao(arq, i, &re);
    if (re.cosN >= 0) {
      if (re.n == 1)
	Pintar(re.xq1, re.yq1, re.xq2, re.yq2, re.xq3, re.yq3,
		 re.xq4, re.yq4, cor_f1_1, cor_int1_1, cor_int2_1, padrao1_1);
      else
	Pintar(re.xq1, re.yq1, re.xq2, re.yq2, re.xq3, re.yq3,
		 re.xq4, re.yq4, cor_f1_2, cor_int1_2, cor_int2_2, padrao1_2);
    } else {
      if (re.n == 1)
	Pintar(re.xq1, re.yq1, re.xq2, re.yq2, re.xq3, re.yq3,
		 re.xq4, re.yq4, cor_f2_1, cor_int3_1, cor_int4_1, padrao2_1);
      else
	Pintar(re.xq1, re.yq1, re.xq2, re.yq2, re.xq3, re.yq3,
		 re.xq4, re.yq4, cor_f2_2, cor_int3_2, cor_int4_2, padrao2_2);
    }
  }
  fclose(arq);

  if (opcao >= 0) {
    if (smax > s1 && smax > s5)
      ArestaDoCubo(xmin, ymin, zmin, xmax, ymin, zmin, opcao);
    if (smax > s1 && smax > s3)
      ArestaDoCubo(xmin, ymin, zmin, xmin, ymax, zmin, opcao);
    if (smax > s7 && smax > s5)
      ArestaDoCubo(xmax, ymax, zmin, xmax, ymin, zmin, opcao);
    if (smax > s7 && smax > s3)
      ArestaDoCubo(xmax, ymax, zmin, xmin, ymax, zmin, opcao);
    if (smax > s2 && smax > s6)
      ArestaDoCubo(xmin, ymin, zmax, xmax, ymin, zmax, opcao);
    if (smax > s2 && smax > s4)
      ArestaDoCubo(xmin, ymin, zmax, xmin, ymax, zmax, opcao);
    if (smax > s8 && smax > s6)
      ArestaDoCubo(xmax, ymax, zmax, xmax, ymin, zmax, opcao);
    if (smax > s8 && smax > s4)
      ArestaDoCubo(xmax, ymax, zmax, xmin, ymax, zmax, opcao);
    if (smax > s1 && smax > s2)
      ArestaDoCubo(xmin, ymin, zmin, xmin, ymin, zmax, opcao);
    if (smax > s5 && smax > s6)
      ArestaDoCubo(xmax, ymin, zmin, xmax, ymin, zmax, opcao);
    if (smax > s3 && smax > s4)
      ArestaDoCubo(xmin, ymax, zmin, xmin, ymax, zmax, opcao);
    if (smax > s7 && smax > s8)
      ArestaDoCubo(xmax, ymax, zmin, xmax, ymax, zmax, opcao);
  }

  remove(arq_dados);

  return 1;
}

/* ------------------------------------------------------------------------- */

/*** FIM DE GRAF3D.C ***/
